@@ -2,7 +2,6 @@ package executor
 
 import (
 	"context"
-	"fmt"
 	"strings"
 	"sync"
 	"testing"
@@ -25,8 +24,8 @@ import (
 
 func TestCoordinator_Execute(t *testing.T) {
 	publishedFalse := overridable.FromBoolOrString(false)
-	srcCLITask := &Task{Repository: testRepo1}
-	sourcegraphTask := &Task{Repository: testRepo2}
+	srcCLITask := &Task{Repository: testRepo1, Steps: []batcheslib.Step{{Run: "echo Hello World"}}}
+	sourcegraphTask := &Task{Repository: testRepo2, Steps: []batcheslib.Step{{Run: "echo Hello Sourcegraph"}}}
 
 	buildSpecFor := func(repo *graphql.Repository, modify func(*batcheslib.ChangesetSpec)) *batcheslib.ChangesetSpec {
 		spec := &batcheslib.ChangesetSpec{
@@ -79,8 +78,8 @@ func TestCoordinator_Execute(t *testing.T) {
 
 			executor: &dummyExecutor{
 				results: []taskResult{
-					{task: srcCLITask, result: execution.Result{Diff: `dummydiff1`}},
-					{task: sourcegraphTask, result: execution.Result{Diff: `dummydiff2`}},
+					{task: srcCLITask, stepResults: []execution.AfterStepResult{{Diff: `dummydiff1`}}},
+					{task: sourcegraphTask, stepResults: []execution.AfterStepResult{{Diff: `dummydiff2`}}},
 				},
 			},
 			opts: NewCoordinatorOpts{Features: featuresAllEnabled()},
@@ -133,7 +132,7 @@ func TestCoordinator_Execute(t *testing.T) {
 				results: []taskResult{
 					{
 						task: srcCLITask,
-						result: execution.Result{
+						stepResults: []execution.AfterStepResult{{
 							Diff: `dummydiff1`,
 							Outputs: map[string]interface{}{
 								"output1": "myOutputValue1",
@@ -141,13 +140,13 @@ func TestCoordinator_Execute(t *testing.T) {
 									"subField": "subFieldValue",
 								},
 							},
-							ChangedFiles: &git.Changes{
+							ChangedFiles: git.Changes{
 								Modified: []string{"modified.txt"},
 								Added:    []string{"added.txt"},
 								Deleted:  []string{"deleted.txt"},
 								Renamed:  []string{"renamed.txt"},
 							},
-						},
+						}},
 					},
 				},
 			},
@@ -198,12 +197,13 @@ func TestCoordinator_Execute(t *testing.T) {
 
 			executor: &dummyExecutor{
 				results: []taskResult{
-					{task: srcCLITask, result: execution.Result{Diff: nestedChangesDiff}},
-					{task: sourcegraphTask, result: execution.Result{Diff: nestedChangesDiff}},
+					{task: srcCLITask, stepResults: []execution.AfterStepResult{{Diff: nestedChangesDiff}}},
+					{task: sourcegraphTask, stepResults: []execution.AfterStepResult{{Diff: nestedChangesDiff}}},
 				},
 			},
 			opts: NewCoordinatorOpts{Features: featuresAllEnabled()},
 
+			// TODO: Fix comment.
 			// We have 4 ChangesetSpecs, but we only want 2 cache entries,
 			// since we cache per Task, not per resulting changeset spec.
 			wantCacheEntries: 2,
@@ -227,8 +227,7 @@ func TestCoordinator_Execute(t *testing.T) {
 			},
 		},
 		{
-			name: "skip errors",
-			opts: NewCoordinatorOpts{Features: featuresAllEnabled(), SkipErrors: true},
+			name: "cache for step mount",
 
 			tasks: []*Task{srcCLITask, sourcegraphTask},
 
@@ -236,29 +235,30 @@ func TestCoordinator_Execute(t *testing.T) {
 				Name:              "my-batch-change",
 				Description:       "the description",
 				ChangesetTemplate: testChangesetTemplate,
+				Steps: []batcheslib.Step{
+					{
+						Run:   "echo foo",
+						Mount: []batcheslib.Mount{{Path: "/foo/bar/sample.sh", Mountpoint: "/tmp/sample.sh"}},
+					},
+				},
 			},
 
-			// Execution succeeded in srcCLIRepo, but fails in sourcegraphRepo
 			executor: &dummyExecutor{
 				results: []taskResult{
-					{task: srcCLITask, result: execution.Result{Diff: `dummydiff1`}},
-				},
-				waitErr: stepFailedErr{
-					Err:         fmt.Errorf("something went wrong"),
-					Run:         "broken command",
-					Container:   "alpine:3",
-					TmpFilename: "/tmp/foobar",
-					Stderr:      "unknown command: broken",
+					{task: srcCLITask, stepResults: []execution.AfterStepResult{{Diff: `dummydiff1`, StepIndex: 0}}},
+					{task: sourcegraphTask, stepResults: []execution.AfterStepResult{{Diff: `dummydiff2`, StepIndex: 0}}},
 				},
 			},
+			opts: NewCoordinatorOpts{Features: featuresAllEnabled()},
 
-			wantErrInclude: "run: broken command",
-			// We want 1 cache entry and 1 spec
-			wantCacheEntries: 1,
+			wantCacheEntries: 2,
 			wantSpecs: []*batcheslib.ChangesetSpec{
 				buildSpecFor(testRepo1, func(spec *batcheslib.ChangesetSpec) {
 					spec.Commits[0].Diff = `dummydiff1`
 				}),
+				buildSpecFor(testRepo2, func(spec *batcheslib.ChangesetSpec) {
+					spec.Commits[0].Diff = `dummydiff2`
+				}),
 			},
 		},
 	}
@@ -276,13 +276,14 @@ func TestCoordinator_Execute(t *testing.T) {
 			}
 
 			logManager := mock.LogNoOpManager{}
+			c := newInMemoryExecutionCache()
+
+			tc.opts.Cache = c
+			tc.opts.Logger = logManager
 
-			cache := newInMemoryExecutionCache()
 			coord := Coordinator{
-				cache:      cache,
-				exec:       tc.executor,
-				logManager: logManager,
-				opts:       tc.opts,
+				exec: tc.executor,
+				opts: tc.opts,
 			}
 
 			// execute contains the actual logic for executing the tasks and
@@ -323,15 +324,15 @@ func TestCoordinator_Execute(t *testing.T) {
 
 			verifyCache := func(t *testing.T) {
 				// Verify that there is a cache entry for each repo.
-				if have, want := cache.size(), tc.wantCacheEntries; have != want {
-					t.Errorf("unexpected number of cache entries: have=%d want=%d cache=%+v", have, want, cache)
+				if have, want := c.size(), tc.wantCacheEntries; have != want {
+					t.Errorf("unexpected number of cache entries: have=%d want=%d cache=%+v", have, want, c)
 				}
 			}
 
 			// Sanity check, since we're going to be looking at the side effects
 			// on the cache.
-			if cache.size() != 0 {
-				t.Fatalf("unexpectedly hot cache: %+v", cache)
+			if c.size() != 0 {
+				t.Fatalf("unexpected hot cache: %+v", c)
 			}
 
 			// Run with a cold cache.
@@ -367,12 +368,6 @@ func TestCoordinator_Execute_StepCaching(t *testing.T) {
 	executor := &dummyExecutor{}
 	executor.results = []taskResult{{
 		task: task,
-		result: execution.Result{
-			Diff:         "dummydiff",
-			ChangedFiles: &git.Changes{},
-			Outputs:      map[string]interface{}{},
-			Path:         "",
-		},
 		stepResults: []execution.AfterStepResult{
 			{StepIndex: 0, Diff: `step-0-diff`},
 			{StepIndex: 1, Diff: `step-1-diff`},
@@ -381,55 +376,59 @@ func TestCoordinator_Execute_StepCaching(t *testing.T) {
 	}}
 
 	// Build Coordinator
-	coord := &Coordinator{cache: cache, exec: executor, logManager: logManager}
+	coord := &Coordinator{
+		opts: NewCoordinatorOpts{
+			Cache:  cache,
+			Logger: logManager,
+		},
+		exec: executor,
+	}
 
 	batchSpec := &batcheslib.BatchSpec{ChangesetTemplate: testChangesetTemplate}
 
 	// First execution. Make sure that the Task executes all steps.
 	execAndEnsure(t, coord, executor, batchSpec, task, assertNoCachedResult(t))
-	// We now expect the cache to have 1+N entries: 1 for the complete task, N
-	// for the steps.
-	wantCacheSize := len(task.Steps) + 1
-	assertCacheSize(t, cache, wantCacheSize)
+	// We now expect the cache to have N entries for N steps.
+	assertCacheSize(t, cache, 3)
 
 	// Reset task
-	task.CachedResultFound = false
+	task.CachedStepResultFound = false
 
 	// Change the 2nd step's definition:
 	task.Steps[1].Run = `echo "two modified"`
 	// Re-execution should start with the diff produced by steps[0] as the
 	// start state from which steps[1] is then re-executed.
-	execAndEnsure(t, coord, executor, batchSpec, task, assertCachedResultForStep(t, 0))
+	execAndEnsure(t, coord, executor, batchSpec, task, func(ctx context.Context, t []*Task, teu TaskExecutionUI) {})
 	// Cache now contains old entries, plus another "complete task" entry and
 	// two entries for newly executed steps.
-	wantCacheSize += 1 + 2
-	assertCacheSize(t, cache, wantCacheSize)
+	assertCacheSize(t, cache, 5)
 
 	// Reset task
-	task.CachedResultFound = false
+	task.CachedStepResultFound = false
 
 	// Change the 3rd step's definition:
 	task.Steps[2].Run = `echo "three modified"`
 	// Re-execution should use the diff from steps[1] as start state
-	execAndEnsure(t, coord, executor, batchSpec, task, assertCachedResultForStep(t, 1))
+	execAndEnsure(t, coord, executor, batchSpec, task, func(ctx context.Context, t []*Task, teu TaskExecutionUI) {})
 	// Cache now contains old entries, plus another "complete task" entry and
 	// a single new step entry
-	wantCacheSize += 1 + 1
-	assertCacheSize(t, cache, wantCacheSize)
+	assertCacheSize(t, cache, 6)
 
 	// Reset task
-	task.CachedResultFound = false
+	task.CachedStepResultFound = false
 
 	// Now we execute the spec with -clear-cache:
 	if err := coord.ClearCache(context.Background(), []*Task{task}); err != nil {
 		t.Fatal(err)
 	}
+	assertCacheSize(t, cache, 3)
+
 	// We don't want any cached results set on the task:
 	execAndEnsure(t, coord, executor, batchSpec, task, assertNoCachedResult(t))
 	// Cache should have the same number of entries: the cached step results should
 	// have been cleared (the complete-task-result is cleared in another
 	// code path) and the same amount of cached entries has been added.
-	assertCacheSize(t, cache, wantCacheSize)
+	assertCacheSize(t, cache, 6)
 }
 
 // execAndEnsure executes the given Task with the given cache and dummyExecutor
@@ -475,24 +474,6 @@ func assertCacheSize(t *testing.T, cache *inMemoryExecutionCache, want int) {
 	}
 }
 
-// assertCachedResultForStep returns a function that can be used as a
-// startCallback on dummyExecutor to assert that the first Task has a cached
-// result for the given step.
-func assertCachedResultForStep(t *testing.T, step int) func(context.Context, []*Task, TaskExecutionUI) {
-	return func(c context.Context, tasks []*Task, ui TaskExecutionUI) {
-		t.Helper()
-
-		task := tasks[0]
-		if !task.CachedResultFound {
-			t.Fatalf("CachedResultFound not set")
-		}
-
-		if have, want := task.CachedResult.StepIndex, step; have != want {
-			t.Fatalf("CachedResult.Step wrong. have=%d, want=%d", have, want)
-		}
-	}
-}
-
 // expectCachedResultForStep returns a function that can be used as a
 // startCallback on dummyExecutor to assert that the first Task has no cached results.
 func assertNoCachedResult(t *testing.T) func(context.Context, []*Task, TaskExecutionUI) {
@@ -500,7 +481,7 @@ func assertNoCachedResult(t *testing.T) func(context.Context, []*Task, TaskExecu
 		t.Helper()
 
 		task := tasks[0]
-		if task.CachedResultFound {
+		if task.CachedStepResultFound {
 			t.Fatalf("CachedResultFound but not expected")
 		}
 	}
@@ -603,7 +584,7 @@ func (c *inMemoryExecutionCache) size() int {
 func (c *inMemoryExecutionCache) getCacheItem(key cache.Keyer) (interface{}, bool, error) {
 	k, err := key.Key()
 	if err != nil {
-		return execution.Result{}, false, err
+		return execution.AfterStepResult{}, false, err
 	}
 
 	c.mu.RLock()
@@ -613,30 +594,7 @@ func (c *inMemoryExecutionCache) getCacheItem(key cache.Keyer) (interface{}, boo
 	return res, ok, nil
 }
 
-func (c *inMemoryExecutionCache) Get(ctx context.Context, key cache.Keyer) (execution.Result, bool, error) {
-	res, ok, err := c.getCacheItem(key)
-	if err != nil || !ok {
-		return execution.Result{}, ok, err
-	}
-
-	execResult, ok := res.(execution.Result)
-	return execResult, ok, nil
-}
-
-func (c *inMemoryExecutionCache) Set(ctx context.Context, key cache.Keyer, result execution.Result) error {
-	k, err := key.Key()
-	if err != nil {
-		return err
-	}
-
-	c.mu.Lock()
-	defer c.mu.Unlock()
-
-	c.cache[k] = result
-	return nil
-}
-
-func (c *inMemoryExecutionCache) GetStepResult(ctx context.Context, key cache.Keyer) (execution.AfterStepResult, bool, error) {
+func (c *inMemoryExecutionCache) Get(ctx context.Context, key cache.Keyer) (execution.AfterStepResult, bool, error) {
 	res, ok, err := c.getCacheItem(key)
 	if err != nil || !ok {
 		return execution.AfterStepResult{}, ok, err
@@ -646,7 +604,7 @@ func (c *inMemoryExecutionCache) GetStepResult(ctx context.Context, key cache.Ke
 	return execResult, ok, nil
 }
 
-func (c *inMemoryExecutionCache) SetStepResult(ctx context.Context, key cache.Keyer, result execution.AfterStepResult) error {
+func (c *inMemoryExecutionCache) Set(ctx context.Context, key cache.Keyer, result execution.AfterStepResult) error {
 	k, err := key.Key()
 	if err != nil {
 		return err
