@@ -13,9 +13,11 @@ import (
 	"testing"
 	"time"
 
-	"github.com/cockroachdb/errors"
 	"github.com/google/go-cmp/cmp"
 	"github.com/sourcegraph/go-diff/diff"
+	"github.com/stretchr/testify/require"
+
+	"github.com/sourcegraph/sourcegraph/lib/errors"
 
 	batcheslib "github.com/sourcegraph/sourcegraph/lib/batches"
 	"github.com/sourcegraph/sourcegraph/lib/batches/execution"
@@ -46,6 +48,12 @@ func TestExecutor_Integration(t *testing.T) {
 	type filesByPath map[string][]string
 	type filesByRepository map[string]filesByPath
 
+	// create a temp directory with a simple shell file
+	tempDir := t.TempDir()
+	mountScript := fmt.Sprintf("%s/sample.sh", tempDir)
+	err := os.WriteFile(mountScript, []byte(`echo -e "foobar\n" >> README.md`), 0777)
+	require.NoError(t, err)
+
 	tests := []struct {
 		name string
 
@@ -69,6 +77,8 @@ func TestExecutor_Integration(t *testing.T) {
 
 		wantFinished        int
 		wantFinishedWithErr int
+
+		wantCacheCount int
 	}{
 		{
 			name: "success",
@@ -97,7 +107,8 @@ func TestExecutor_Integration(t *testing.T) {
 					rootPath: []string{"README.md"},
 				},
 			},
-			wantFinished: 2,
+			wantFinished:   2,
+			wantCacheCount: 4,
 		},
 		{
 			name: "empty",
@@ -120,7 +131,8 @@ func TestExecutor_Integration(t *testing.T) {
 					rootPath: []string{},
 				},
 			},
-			wantFinished: 1,
+			wantFinished:   1,
+			wantCacheCount: 1,
 		},
 		{
 			name: "timeout",
@@ -178,7 +190,8 @@ func TestExecutor_Integration(t *testing.T) {
 					},
 				},
 			},
-			wantFinished: 1,
+			wantFinished:   1,
+			wantCacheCount: 5,
 		},
 		{
 			name: "workspaces",
@@ -233,7 +246,8 @@ func TestExecutor_Integration(t *testing.T) {
 					"a/b":    []string{"a/b/hello.txt", "a/b/gitignore-exists", "a/b/gitignore-exists-in-a"},
 				},
 			},
-			wantFinished: 3,
+			wantFinished:   3,
+			wantCacheCount: 15,
 		},
 		{
 			name: "step condition",
@@ -268,7 +282,8 @@ func TestExecutor_Integration(t *testing.T) {
 					"sub/directory/of/repo": []string{"README.md", "hello.txt", "in-path.txt"},
 				},
 			},
-			wantFinished: 2,
+			wantFinished:   2,
+			wantCacheCount: 4,
 		},
 		{
 			name: "skips errors",
@@ -300,6 +315,31 @@ func TestExecutor_Integration(t *testing.T) {
 			wantErrInclude:      "execution in github.com/sourcegraph/sourcegraph failed: run: exit 1",
 			wantFinished:        1,
 			wantFinishedWithErr: 1,
+			wantCacheCount:      2,
+		},
+		{
+			name: "mount path",
+			archives: []mock.RepoArchive{
+				{RepoName: testRepo1.Name, Commit: testRepo1.Rev(), Files: map[string]string{
+					"README.md": "# Welcome to the README\n",
+				}},
+			},
+			steps: []batcheslib.Step{
+				{
+					Run:   mountScript,
+					Mount: []batcheslib.Mount{{Path: mountScript, Mountpoint: mountScript}},
+				},
+			},
+			tasks: []*Task{
+				{Repository: testRepo1},
+			},
+			wantFilesChanged: filesByRepository{
+				testRepo1.ID: filesByPath{
+					rootPath: []string{"README.md"},
+				},
+			},
+			wantFinished:   1,
+			wantCacheCount: 1,
 		},
 	}
 
@@ -335,15 +375,12 @@ func TestExecutor_Integration(t *testing.T) {
 			client := api.NewClient(api.ClientOpts{Endpoint: ts.URL, Out: &clientBuffer})
 
 			// Temp dir for log files and downloaded archives
-			testTempDir, err := os.MkdirTemp("", "executor-integration-test-*")
-			if err != nil {
-				t.Fatal(err)
-			}
-			defer os.Remove(testTempDir)
+			testTempDir := t.TempDir()
 
+			cr, _ := workspace.NewCreator(context.Background(), "bind", testTempDir, testTempDir, images)
 			// Setup executor
-			opts := newExecutorOpts{
-				Creator:             workspace.NewCreator(context.Background(), "bind", testTempDir, testTempDir, images),
+			opts := NewExecutorOpts{
+				Creator:             cr,
 				RepoArchiveRegistry: repozip.NewArchiveRegistry(client, testTempDir, false),
 				Logger:              mock.LogNoOpManager{},
 				EnsureImage:         imageMapEnsurer(images),
@@ -358,7 +395,7 @@ func TestExecutor_Integration(t *testing.T) {
 			}
 
 			dummyUI := newDummyTaskExecutionUI()
-			executor := newExecutor(opts)
+			executor := NewExecutor(opts)
 
 			// Run executor
 			executor.Start(context.Background(), tc.tasks, dummyUI)
@@ -386,11 +423,22 @@ func TestExecutor_Integration(t *testing.T) {
 				}
 			}
 
-			if have, want := len(results), wantResults; have != want {
-				t.Fatalf("wrong number of execution results. want=%d, have=%d", want, have)
+			haveResults := 0
+			for _, res := range results {
+				if res.err == nil {
+					haveResults++
+				}
+			}
+
+			if have, want := haveResults, wantResults; have != want {
+				t.Fatalf("wrong number of results. want=%d, have=%d", want, have)
 			}
 
 			for _, taskResult := range results {
+				if taskResult.err != nil {
+					continue
+				}
+
 				repoID := taskResult.task.Repository.ID
 				path := taskResult.task.Path
 
@@ -406,7 +454,9 @@ func TestExecutor_Integration(t *testing.T) {
 					t.Fatalf("spec for repo %q and path %q but no files expected in that branch", repoID, path)
 				}
 
-				fileDiffs, err := diff.ParseMultiFileDiff([]byte(taskResult.result.Diff))
+				lastStepResult := taskResult.stepResults[len(taskResult.stepResults)-1]
+
+				fileDiffs, err := diff.ParseMultiFileDiff([]byte(lastStepResult.Diff))
 				if err != nil {
 					t.Fatalf("failed to parse diff: %s", err)
 				}
@@ -494,12 +544,11 @@ index 02a19af..c9644dd 100644
 			Steps: []batcheslib.Step{
 				{Run: `echo -e "foobar\n" >> README.md`},
 			},
-			CachedResultFound: true,
-			CachedResult: execution.AfterStepResult{
-				StepIndex:          0,
-				Diff:               cachedDiff,
-				Outputs:            map[string]interface{}{},
-				PreviousStepResult: execution.StepResult{},
+			CachedStepResultFound: true,
+			CachedStepResult: execution.AfterStepResult{
+				StepIndex: 0,
+				Diff:      cachedDiff,
+				Outputs:   map[string]interface{}{},
 			},
 			Repository: testRepo1,
 		}
@@ -510,22 +559,22 @@ index 02a19af..c9644dd 100644
 		}
 
 		if have, want := len(results), 1; have != want {
-			t.Fatalf("wrong number of execution results. want=%d, have=%d", want, have)
-		}
-
-		// We want the diff to be the same as the cached one, since we only had to
-		// execute a single step
-		executionResult := results[0].result
-		if diff := cmp.Diff(executionResult.Diff, cachedDiff); diff != "" {
-			t.Fatalf("wrong diff: %s", diff)
+			t.Fatalf("wrong number of results. want=%d, have=%d", want, have)
 		}
 
 		if have, want := len(results[0].stepResults), 1; have != want {
 			t.Fatalf("wrong length of step results. have=%d, want=%d", have, want)
 		}
 
+		// We want the diff to be the same as the cached one, since we only had to
+		// execute a single step
+		executionResult := results[0].stepResults[0]
+		if diff := cmp.Diff(executionResult.Diff, cachedDiff); diff != "" {
+			t.Fatalf("wrong diff: %s", diff)
+		}
+
 		stepResult := results[0].stepResults[0]
-		if diff := cmp.Diff(stepResult, task.CachedResult); diff != "" {
+		if diff := cmp.Diff(stepResult, task.CachedStepResult); diff != "" {
 			t.Fatalf("wrong stepResult: %s", diff)
 		}
 	})
@@ -610,21 +659,19 @@ echo "previous_step.modified_files=${{ previous_step.modified_files }}" >> READM
 `},
 				{Run: `echo "this is step 5" >> ${{ outputs.myOutput }}`},
 			},
-			CachedResultFound: true,
-			CachedResult: execution.AfterStepResult{
+			CachedStepResultFound: true,
+			CachedStepResult: execution.AfterStepResult{
 				StepIndex: 2,
 				Diff:      cachedDiff,
 				Outputs: map[string]interface{}{
 					"myOutput": "my-output.txt",
 				},
-				PreviousStepResult: execution.StepResult{
-					Files: &git.Changes{
-						Modified: []string{"README.md"},
-						Added:    []string{"README.txt"},
-					},
-					Stdout: nil,
-					Stderr: nil,
+				ChangedFiles: git.Changes{
+					Modified: []string{"README.md"},
+					Added:    []string{"README.txt"},
 				},
+				Stdout: "",
+				Stderr: "",
 			},
 		}
 
@@ -634,15 +681,15 @@ echo "previous_step.modified_files=${{ previous_step.modified_files }}" >> READM
 		}
 
 		if have, want := len(results), 1; have != want {
-			t.Fatalf("wrong number of execution results. want=%d, have=%d", want, have)
+			t.Fatalf("wrong number of results. want=%d, have=%d", want, have)
 		}
 
-		executionResult := results[0].result
+		executionResult := results[0].stepResults[len(results[0].stepResults)-1]
 		if diff := cmp.Diff(executionResult.Diff, wantFinalDiff); diff != "" {
 			t.Fatalf("wrong diff: %s", diff)
 		}
 
-		if diff := cmp.Diff(executionResult.Outputs, task.CachedResult.Outputs); diff != "" {
+		if diff := cmp.Diff(executionResult.Outputs, task.CachedStepResult.Outputs); diff != "" {
 			t.Fatalf("wrong execution result outputs: %s", diff)
 		}
 
@@ -656,7 +703,78 @@ echo "previous_step.modified_files=${{ previous_step.modified_files }}" >> READM
 			t.Fatalf("wrong stepIndex. have=%d, want=%d", have, want)
 		}
 
-		if diff := cmp.Diff(lastStepResult.Outputs, task.CachedResult.Outputs); diff != "" {
+		if diff := cmp.Diff(lastStepResult.Outputs, task.CachedStepResult.Outputs); diff != "" {
+			t.Fatalf("wrong step result outputs: %s", diff)
+		}
+	})
+
+	t.Run("step stdout cached", func(t *testing.T) {
+		archive := mock.RepoArchive{
+			RepoName: testRepo1.Name, Commit: testRepo1.Rev(),
+			Files: map[string]string{
+				"README.md": `# automation-testing
+This repository is used to test opening and closing pull request with Automation
+`,
+			},
+		}
+
+		wantFinalDiff := `diff --git README.md README.md
+index 3040106..5f2f924 100644
+--- README.md
++++ README.md
+@@ -1,2 +1,3 @@
+ # automation-testing
+ This repository is used to test opening and closing pull request with Automation
++hello world
+`
+
+		task := &Task{
+			Repository:            testRepo1,
+			BatchChangeAttributes: &template.BatchChangeAttributes{},
+			Steps: []batcheslib.Step{
+				{Run: "echo -n Hello world"},
+				{Run: `echo ${{ previous_step.stdout }} >> README.md`},
+			},
+			CachedStepResultFound: true,
+			CachedStepResult: execution.AfterStepResult{
+				StepIndex:    0,
+				Diff:         "",
+				Outputs:      map[string]interface{}{},
+				ChangedFiles: git.Changes{},
+				Stdout:       "hello world",
+				Stderr:       "",
+			},
+		}
+
+		results, err := testExecuteTasks(t, []*Task{task}, archive)
+		if err != nil {
+			t.Fatalf("execution failed: %s", err)
+		}
+
+		if have, want := len(results), 1; have != want {
+			t.Fatalf("wrong number of results. want=%d, have=%d", want, have)
+		}
+
+		executionResult := results[0].stepResults[len(results[0].stepResults)-1]
+		if diff := cmp.Diff(executionResult.Diff, wantFinalDiff); diff != "" {
+			t.Fatalf("wrong diff: %s", diff)
+		}
+
+		if diff := cmp.Diff(executionResult.Outputs, task.CachedStepResult.Outputs); diff != "" {
+			t.Fatalf("wrong execution result outputs: %s", diff)
+		}
+
+		// Only one step should've been executed
+		if have, want := len(results[0].stepResults), 1; have != want {
+			t.Fatalf("wrong length of step results. have=%d, want=%d", have, want)
+		}
+
+		lastStepResult := results[0].stepResults[0]
+		if have, want := lastStepResult.StepIndex, 1; have != want {
+			t.Fatalf("wrong stepIndex. have=%d, want=%d", have, want)
+		}
+
+		if diff := cmp.Diff(lastStepResult.Outputs, task.CachedStepResult.Outputs); diff != "" {
 			t.Fatalf("wrong step result outputs: %s", diff)
 		}
 	})
@@ -667,11 +785,7 @@ func testExecuteTasks(t *testing.T, tasks []*Task, archives ...mock.RepoArchive)
 		t.Skip("Test doesn't work on Windows because dummydocker is written in bash")
 	}
 
-	testTempDir, err := os.MkdirTemp("", "executor-integration-test-*")
-	if err != nil {
-		t.Fatal(err)
-	}
-	t.Cleanup(func() { os.Remove(testTempDir) })
+	testTempDir := t.TempDir()
 
 	// Setup dummydocker
 	addToPath(t, "testdata/dummydocker")
@@ -693,9 +807,10 @@ func testExecuteTasks(t *testing.T, tasks []*Task, archives ...mock.RepoArchive)
 		}
 	}
 
+	cr, _ := workspace.NewCreator(context.Background(), "bind", testTempDir, testTempDir, images)
 	// Setup executor
-	executor := newExecutor(newExecutorOpts{
-		Creator:             workspace.NewCreator(context.Background(), "bind", testTempDir, testTempDir, images),
+	executor := NewExecutor(NewExecutorOpts{
+		Creator:             cr,
 		RepoArchiveRegistry: repozip.NewArchiveRegistry(client, testTempDir, false),
 		Logger:              mock.LogNoOpManager{},
 		EnsureImage:         imageMapEnsurer(images),
