@@ -11,15 +11,14 @@ import (
 	"strings"
 	"time"
 
-	"github.com/cockroachdb/errors"
-	"github.com/hashicorp/go-multierror"
-
 	batcheslib "github.com/sourcegraph/sourcegraph/lib/batches"
 	"github.com/sourcegraph/sourcegraph/lib/batches/execution"
 	"github.com/sourcegraph/sourcegraph/lib/batches/git"
 	"github.com/sourcegraph/sourcegraph/lib/batches/template"
+	"github.com/sourcegraph/sourcegraph/lib/errors"
 
 	"github.com/sourcegraph/src-cli/internal/batches/log"
+	"github.com/sourcegraph/src-cli/internal/batches/repozip"
 	"github.com/sourcegraph/src-cli/internal/batches/util"
 	"github.com/sourcegraph/src-cli/internal/batches/workspace"
 
@@ -28,121 +27,145 @@ import (
 	yamlv3 "gopkg.in/yaml.v3"
 )
 
-type executionOpts struct {
-	wc          workspace.Creator
-	ensureImage imageEnsurer
-
-	task *Task
-
-	tempDir string
-
-	logger log.TaskLogger
-
-	ui StepsExecutionUI
+type RunStepsOpts struct {
+	// WC is the workspace creator to use. It will be called at the beginning of
+	// runSteps to prepare the workspace for execution.
+	WC workspace.Creator
+	// EnsureImage is called in runSteps to make sure the used image has been
+	// pulled from the registry.
+	EnsureImage imageEnsurer
+	// Task is the definition of the workspace execution.
+	Task *Task
+	// TempDir points to where temporary files of the execution should live at.
+	TempDir string
+	// Timeout sets the deadline for the execution context. When exceeded,
+	// execution will stop and an error is returned.
+	Timeout time.Duration
+	// RepoArchive is the repo archive to be used for creating the workspace.
+	RepoArchive repozip.Archive
+	Logger      log.TaskLogger
+	UI          StepsExecutionUI
+	// IsRemote toggles server-side execution mode. This disables file mounts using
+	// step.mounts.
+	IsRemote bool
+	// GlobalEnv is the os.Environ() for the execution. We don't read from os.Environ()
+	// directly to allow injecting variables and hiding others.
+	GlobalEnv []string
 }
 
-func runSteps(ctx context.Context, opts *executionOpts) (result execution.Result, stepResults []execution.AfterStepResult, err error) {
-	opts.ui.ArchiveDownloadStarted()
-	err = opts.task.Archive.Ensure(ctx)
-	opts.ui.ArchiveDownloadFinished(err)
-	if err != nil {
-		return execution.Result{}, nil, errors.Wrap(err, "fetching repo")
-	}
-	defer opts.task.Archive.Close()
+func RunSteps(ctx context.Context, opts *RunStepsOpts) (stepResults []execution.AfterStepResult, err error) {
+	// Set up our timeout.
+	ctx, cancel := context.WithTimeout(ctx, opts.Timeout)
+	defer cancel()
 
-	opts.ui.WorkspaceInitializationStarted()
-	workspace, err := opts.wc.Create(ctx, opts.task.Repository, opts.task.Steps, opts.task.Archive)
+	// Return an errTimeoutReached error in case the deadline has been exceeded.
+	defer func() {
+		if err != nil {
+			if reachedTimeout(ctx, err) {
+				err = &errTimeoutReached{timeout: opts.Timeout}
+			}
+		}
+	}()
+
+	opts.UI.ArchiveDownloadStarted()
+	err = opts.RepoArchive.Ensure(ctx)
+	opts.UI.ArchiveDownloadFinished(err)
 	if err != nil {
-		return execution.Result{}, nil, errors.Wrap(err, "creating workspace")
+		return nil, errors.Wrap(err, "fetching repo")
 	}
-	defer workspace.Close(ctx)
-	opts.ui.WorkspaceInitializationFinished()
+	defer opts.RepoArchive.Close()
+
+	opts.UI.WorkspaceInitializationStarted()
+	ws, err := opts.WC.Create(ctx, opts.Task.Repository, opts.Task.Steps, opts.RepoArchive)
+	if err != nil {
+		return nil, errors.Wrap(err, "creating workspace")
+	}
+	defer ws.Close(ctx)
+	opts.UI.WorkspaceInitializationFinished()
 
 	var (
-		execResult = execution.Result{
-			Diff:         "",
-			ChangedFiles: &git.Changes{},
-			Outputs:      make(map[string]interface{}),
-			Path:         opts.task.Path,
-		}
-		previousStepResult execution.StepResult
-		startStep          int
+		// lastOutputs are the step.outputs from the previously run step.
+		// Outputs are additive, so we will only ever append to this map,
+		// but never clear it.
+		lastOutputs        = make(map[string]any)
+		previousStepResult execution.AfterStepResult
+		// startStep is the index of the step we start executing with. Set when
+		// there's a partial cache result for the task.
+		startStep int
 	)
 
-	if opts.task.CachedResultFound {
+	if opts.Task.CachedStepResultFound {
 		// Set the Outputs to the cached outputs
-		execResult.Outputs = opts.task.CachedResult.Outputs
-
-		lastStep := opts.task.CachedResult.StepIndex
+		lastOutputs = opts.Task.CachedStepResult.Outputs
+		previousStepResult = opts.Task.CachedStepResult
+		lastStep := opts.Task.CachedStepResult.StepIndex
+		// Add the cached result to the result set. It may be required for building
+		// the changeset specs in the end if we don't need to execute any other steps
+		// due to dynamic skipping.
+		// We remove it at the end of this function if it isn't needed, so we don't
+		// write it to the cache again.
+		stepResults = append(stepResults, previousStepResult)
 
 		// If we have cached results and don't need to execute any more steps,
-		// we can quit
-		if lastStep == len(opts.task.Steps)-1 {
-			changes, err := git.ChangesInDiff([]byte(opts.task.CachedResult.Diff))
-			if err != nil {
-				return execResult, nil, errors.Wrap(err, "parsing cached step diff")
+		// we can quit.
+		// TODO: This doesn't consider skipped steps, but it's not _that_
+		// bad, we will find out about that further down and correctly finish
+		// execution early.
+		if lastStep == len(opts.Task.Steps)-1 {
+			return stepResults, nil
+		}
+
+		// If the previous steps made any modifications to the workspace yet,
+		// apply them.
+		if opts.Task.CachedStepResult.Diff != "" {
+			if err := ws.ApplyDiff(ctx, []byte(opts.Task.CachedStepResult.Diff)); err != nil {
+				return nil, errors.Wrap(err, "applying diff of cache result")
 			}
-
-			execResult.Diff = opts.task.CachedResult.Diff
-			execResult.ChangedFiles = &changes
-			stepResults = append(stepResults, opts.task.CachedResult)
-
-			return execResult, stepResults, nil
 		}
 
 		startStep = lastStep + 1
 
-		opts.ui.SkippingStepsUpto(
+		opts.UI.SkippingStepsUpto(
 			// UI is 1-indexed.
 			startStep + 1,
 		)
 	}
 
-	for i := startStep; i < len(opts.task.Steps); i++ {
-		step := opts.task.Steps[i]
+	for i := startStep; i < len(opts.Task.Steps); i++ {
+		step := opts.Task.Steps[i]
 
 		stepContext := template.StepContext{
-			BatchChange: *opts.task.BatchChangeAttributes,
-			Repository:  util.NewTemplatingRepo(opts.task.Repository.Name, opts.task.Repository.FileMatches),
-			Outputs:     execResult.Outputs,
+			BatchChange: *opts.Task.BatchChangeAttributes,
+			Repository:  util.NewTemplatingRepo(opts.Task.Repository.Name, opts.Task.Repository.FileMatches),
+			Outputs:     lastOutputs,
 			Steps: template.StepsContext{
-				Path:    execResult.Path,
-				Changes: previousStepResult.Files,
+				Path:    opts.Task.Path,
+				Changes: previousStepResult.ChangedFiles,
 			},
 			PreviousStep: previousStepResult,
 		}
 
-		if opts.task.CachedResultFound && i == startStep {
-			previousStepResult = opts.task.CachedResult.PreviousStepResult
-
-			stepContext.PreviousStep = previousStepResult
-			stepContext.Steps.Changes = previousStepResult.Files
-			stepContext.Outputs = opts.task.CachedResult.Outputs
-
-			if err := workspace.ApplyDiff(ctx, []byte(opts.task.CachedResult.Diff)); err != nil {
-				return execResult, nil, errors.Wrap(err, "getting changed files in step")
-			}
-		}
-
+		// Check if the step needs to be skipped.
 		cond, err := template.EvalStepCondition(step.IfCondition(), &stepContext)
 		if err != nil {
-			return execResult, nil, errors.Wrap(err, "evaluating step condition")
+			return nil, errors.Wrap(err, "evaluating step condition")
 		}
 		if !cond {
-			opts.ui.StepSkipped(i + 1)
+			opts.UI.StepSkipped(i + 1)
 			continue
 		}
 
 		// We need to grab the digest for the exact image we're using.
-		img, err := opts.ensureImage(ctx, step.Container)
+		img, err := opts.EnsureImage(ctx, step.Container)
 		if err != nil {
-			return execResult, nil, err
+			return nil, err
 		}
 		digest, err := img.Digest(ctx)
 		if err != nil {
-			return execResult, nil, err
+			return nil, err
 		}
-		stdoutBuffer, stderrBuffer, err := executeSingleStep(ctx, opts, workspace, i, step, digest, &stepContext)
+
+		stdoutBuffer, stderrBuffer, err := executeSingleStep(ctx, opts, ws, i, step, digest, &stepContext)
 		defer func() {
 			if err != nil {
 				exitCode := -1
@@ -150,80 +173,80 @@ func runSteps(ctx context.Context, opts *executionOpts) (result execution.Result
 				if errors.As(err, sfe) {
 					exitCode = sfe.ExitCode
 				}
-				opts.ui.StepFailed(i+1, err, exitCode)
+				opts.UI.StepFailed(i+1, err, exitCode)
 			}
 		}()
 		if err != nil {
-			return execResult, nil, err
-		}
-
-		changes, err := workspace.Changes(ctx)
-		if err != nil {
-			return execResult, nil, errors.Wrap(err, "getting changed files in step")
-		}
-
-		result := execution.StepResult{Files: changes, Stdout: &stdoutBuffer, Stderr: &stderrBuffer}
-
-		// Set stepContext.Step to current step's results before rendering outputs
-		stepContext.Step = result
-		// Render and evaluate outputs
-		if err := setOutputs(step.Outputs, execResult.Outputs, &stepContext); err != nil {
-			return execResult, nil, errors.Wrap(err, "setting step outputs")
+			return stepResults, err
 		}
 
 		// Get the current diff and store that away as the per-step result.
-		stepDiff, err := workspace.Diff(ctx)
+		stepDiff, err := ws.Diff(ctx)
 		if err != nil {
-			return execResult, nil, errors.Wrap(err, "getting diff produced by step")
+			return stepResults, errors.Wrap(err, "getting diff produced by step")
 		}
+
+		// Next parse the diff to determine which files were changed.
+		changes, err := git.ChangesInDiff(stepDiff)
+		if err != nil {
+			return stepResults, errors.Wrap(err, "getting changed files in step")
+		}
+
 		stepResult := execution.AfterStepResult{
-			StepIndex:          i,
-			Diff:               string(stepDiff),
-			Outputs:            make(map[string]interface{}),
-			PreviousStepResult: stepContext.PreviousStep,
+			ChangedFiles: changes,
+			Stdout:       stdoutBuffer.String(),
+			Stderr:       stderrBuffer.String(),
+			StepIndex:    i,
+			Diff:         string(stepDiff),
+			// Those will be set below.
+			Outputs: make(map[string]interface{}),
 		}
-		for k, v := range execResult.Outputs {
+
+		// Set stepContext.Step to current step's results before rendering outputs.
+		stepContext.Step = stepResult
+		// Render and evaluate outputs.
+		if err := setOutputs(step.Outputs, lastOutputs, &stepContext); err != nil {
+			return stepResults, errors.Wrap(err, "setting step outputs")
+		}
+		for k, v := range lastOutputs {
 			stepResult.Outputs[k] = v
 		}
 		stepResults = append(stepResults, stepResult)
-		previousStepResult = result
+		previousStepResult = stepResult
 
-		opts.ui.StepFinished(i+1, stepResult.Diff, result.Files, stepResult.Outputs)
+		opts.UI.StepFinished(i+1, stepResult.Diff, stepResult.ChangedFiles, stepResult.Outputs)
 	}
 
-	opts.ui.CalculatingDiffStarted()
-	diffOut, err := workspace.Diff(ctx)
-	if err != nil {
-		return execResult, nil, errors.Wrap(err, "git diff failed")
+	// If we ended up executing at least one step in addition to the cached step,
+	// we can remove it from the result set. Otherwise we write it to the cache
+	// again which is unnecessary. We only need to return it when we want to build changeset
+	// specs from it.
+	if len(stepResults) > 1 && opts.Task.CachedStepResultFound {
+		stepResults = stepResults[1:]
 	}
 
-	opts.ui.CalculatingDiffFinished()
-
-	execResult.Diff = string(diffOut)
-	execResult.ChangedFiles = previousStepResult.Files
-
-	return execResult, stepResults, err
+	return stepResults, err
 }
 
 const workDir = "/work"
 
 func executeSingleStep(
 	ctx context.Context,
-	opts *executionOpts,
+	opts *RunStepsOpts,
 	workspace workspace.Workspace,
-	i int,
+	stepIdx int,
 	step batcheslib.Step,
 	imageDigest string,
 	stepContext *template.StepContext,
-) (bytes.Buffer, bytes.Buffer, error) {
+) (stdout bytes.Buffer, stderr bytes.Buffer, err error) {
 	// ----------
 	// PREPARATION
 	// ----------
-	opts.ui.StepPreparingStart(i + 1)
+	opts.UI.StepPreparingStart(stepIdx + 1)
 
-	cidFile, cleanup, err := createCidFile(ctx, opts.tempDir, util.SlugForRepo(opts.task.Repository.Name, opts.task.Repository.Rev()))
+	cidFile, cleanup, err := createCidFile(ctx, opts.TempDir, util.SlugForRepo(opts.Task.Repository.Name, opts.Task.Repository.Rev()))
 	if err != nil {
-		opts.ui.StepPreparingFailed(i+1, err)
+		opts.UI.StepPreparingFailed(stepIdx+1, err)
 		return bytes.Buffer{}, bytes.Buffer{}, err
 	}
 	defer cleanup()
@@ -232,46 +255,47 @@ func executeSingleStep(
 	shell, containerTemp, err := probeImageForShell(ctx, imageDigest)
 	if err != nil {
 		err = errors.Wrapf(err, "probing image %q for shell", step.Container)
-		opts.ui.StepPreparingFailed(i+1, err)
+		opts.UI.StepPreparingFailed(stepIdx+1, err)
 		return bytes.Buffer{}, bytes.Buffer{}, err
 	}
 
-	runScriptFile, runScript, cleanup, err := createRunScriptFile(ctx, opts.tempDir, step.Run, stepContext)
+	runScriptFile, runScript, cleanup, err := createRunScriptFile(ctx, opts.TempDir, step.Run, stepContext)
 	if err != nil {
-		opts.ui.StepPreparingFailed(i+1, err)
+		opts.UI.StepPreparingFailed(stepIdx+1, err)
 		return bytes.Buffer{}, bytes.Buffer{}, err
 	}
 	defer cleanup()
 
 	// Parse and render the step.Files.
-	filesToMount, cleanup, err := createFilesToMount(opts.tempDir, step, stepContext)
+	filesToMount, cleanup, err := createFilesToMount(opts.TempDir, step, stepContext)
 	if err != nil {
-		opts.ui.StepPreparingFailed(i+1, err)
+		opts.UI.StepPreparingFailed(stepIdx+1, err)
 		return bytes.Buffer{}, bytes.Buffer{}, err
 	}
 	defer cleanup()
 
 	// Resolve step.Env given the current environment.
-	stepEnv, err := step.Env.Resolve(os.Environ())
+	stepEnv, err := step.Env.Resolve(opts.GlobalEnv)
 	if err != nil {
 		err = errors.Wrap(err, "resolving step environment")
-		opts.ui.StepPreparingFailed(i+1, err)
+		opts.UI.StepPreparingFailed(stepIdx+1, err)
 		return bytes.Buffer{}, bytes.Buffer{}, err
 	}
+
 	// Render the step.Env variables as templates.
 	env, err := template.RenderStepMap(stepEnv, stepContext)
 	if err != nil {
 		err = errors.Wrap(err, "parsing step environment")
-		opts.ui.StepPreparingFailed(i+1, err)
+		opts.UI.StepPreparingFailed(stepIdx+1, err)
 		return bytes.Buffer{}, bytes.Buffer{}, err
 	}
 
-	opts.ui.StepPreparingSuccess(i + 1)
+	opts.UI.StepPreparingSuccess(stepIdx + 1)
 
 	// ----------
 	// EXECUTION
 	// ----------
-	opts.ui.StepStarted(i+1, runScript, env)
+	opts.UI.StepStarted(stepIdx+1, runScript, env)
 
 	workspaceOpts, err := workspace.DockerRunOpts(ctx, workDir)
 	if err != nil {
@@ -280,8 +304,8 @@ func executeSingleStep(
 
 	// Where should we execute the steps.run script?
 	scriptWorkDir := workDir
-	if opts.task.Path != "" {
-		scriptWorkDir = workDir + "/" + opts.task.Path
+	if opts.Task.Path != "" {
+		scriptWorkDir = workDir + "/" + opts.Task.Path
 	}
 
 	args := append([]string{
@@ -297,6 +321,14 @@ func executeSingleStep(
 		args = append(args, "--mount", fmt.Sprintf("type=bind,source=%s,target=%s,ro", source.Name(), target))
 	}
 
+	// Temporarily add a guard to prevent a path to mount path for server-side processing.
+	if !opts.IsRemote {
+		// Mount any paths on the local system to the docker container. The paths have already been validated during parsing.
+		for _, mount := range step.Mount {
+			args = append(args, "--mount", fmt.Sprintf("type=bind,source=%s,target=%s,ro", mount.Path, mount.Mountpoint))
+		}
+	}
+
 	for k, v := range env {
 		args = append(args, "-e", k+"="+v)
 	}
@@ -311,19 +343,18 @@ func executeSingleStep(
 
 	writerCtx, writerCancel := context.WithCancel(ctx)
 	defer writerCancel()
-	outputWriter := opts.ui.StepOutputWriter(writerCtx, opts.task, i+1)
+	outputWriter := opts.UI.StepOutputWriter(writerCtx, opts.Task, stepIdx+1)
 	defer func() {
 		outputWriter.Close()
 	}()
 
-	var stdoutBuffer, stderrBuffer bytes.Buffer
-	stdout := io.MultiWriter(&stdoutBuffer, outputWriter.StdoutWriter(), opts.logger.PrefixWriter("stdout"))
-	stderr := io.MultiWriter(&stderrBuffer, outputWriter.StderrWriter(), opts.logger.PrefixWriter("stderr"))
+	stdoutWriter := io.MultiWriter(&stdout, outputWriter.StdoutWriter(), opts.Logger.PrefixWriter("stdout"))
+	stderrWriter := io.MultiWriter(&stderr, outputWriter.StderrWriter(), opts.Logger.PrefixWriter("stderr"))
 
 	// Setup readers that pipe the output into the given buffers
-	wg, err := process.PipeOutput(ctx, cmd, stdout, stderr)
+	wg, err := process.PipeOutput(ctx, cmd, stdoutWriter, stderrWriter)
 	if err != nil {
-		return bytes.Buffer{}, bytes.Buffer{}, errors.Wrap(err, "piping process output")
+		return stdout, stderr, errors.Wrap(err, "piping process output")
 	}
 
 	newStepFailedErr := func(wrappedErr error) stepFailedErr {
@@ -339,34 +370,35 @@ func executeSingleStep(
 			Run:         runScript,
 			Container:   step.Container,
 			TmpFilename: containerTemp,
-			Stdout:      strings.TrimSpace(stdoutBuffer.String()),
-			Stderr:      strings.TrimSpace(stderrBuffer.String()),
+			Stdout:      strings.TrimSpace(stdout.String()),
+			Stderr:      strings.TrimSpace(stderr.String()),
 		}
 	}
 
-	opts.logger.Logf("[Step %d] run: %q, container: %q", i+1, step.Run, step.Container)
-	opts.logger.Logf("[Step %d] full command: %q", i+1, strings.Join(cmd.Args, " "))
+	opts.Logger.Logf("[Step %d] run: %q, container: %q", stepIdx+1, step.Run, step.Container)
+	opts.Logger.Logf("[Step %d] full command: %q", stepIdx+1, strings.Join(cmd.Args, " "))
 
-	// Start the command
+	// Start the command.
 	t0 := time.Now()
 	if err := cmd.Start(); err != nil {
-		opts.logger.Logf("[Step %d] error starting Docker container: %+v", i+1, err)
-		return stdoutBuffer, stderrBuffer, newStepFailedErr(err)
+		opts.Logger.Logf("[Step %d] error starting Docker container: %+v", stepIdx+1, err)
+		return stdout, stderr, newStepFailedErr(err)
 	}
 
 	// Wait for the readers, because the pipes used by PipeOutput under the
-	// hood are closed when the command exits
+	// hood are closed when the command exits.
 	wg.Wait()
-	// Now wait for the command
+
+	// Now wait for the command.
 	err = cmd.Wait()
 	elapsed := time.Since(t0).Round(time.Millisecond)
 	if err != nil {
-		opts.logger.Logf("[Step %d] took %s; error running Docker container: %+v", i+1, elapsed, err)
-		return stdoutBuffer, stderrBuffer, newStepFailedErr(err)
+		opts.Logger.Logf("[Step %d] took %s; error running Docker container: %+v", stepIdx+1, elapsed, err)
+		return stdout, stderr, newStepFailedErr(err)
 	}
 
-	opts.logger.Logf("[Step %d] complete in %s", i+1, elapsed)
-	return stdoutBuffer, stderrBuffer, nil
+	opts.Logger.Logf("[Step %d] complete in %s", stepIdx+1, elapsed)
+	return stdout, stderr, nil
 }
 
 func setOutputs(stepOutputs batcheslib.Outputs, global map[string]interface{}, stepCtx *template.StepContext) error {
@@ -417,11 +449,7 @@ func probeImageForShell(ctx context.Context, image string) (shell, tempfile stri
 	// same time by trying to run /bin/bash -c mktemp,
 	// followed by /bin/sh -c mktemp.
 
-	// We'll also set up our error.
-	err = new(multierror.Error)
-
-	// Now we can iterate through our shell options and try to run mktemp with
-	// them.
+	// We can iterate through our shell options and try to run mktemp with them.
 	for _, shell = range []string{"/bin/bash", "/bin/sh"} {
 		stdout := new(bytes.Buffer)
 		stderr := new(bytes.Buffer)
@@ -433,7 +461,7 @@ func probeImageForShell(ctx context.Context, image string) (shell, tempfile stri
 		cmd.Stderr = stderr
 
 		if runErr := cmd.Run(); runErr != nil {
-			err = multierror.Append(err, errors.Wrapf(runErr, "probing shell %q:\n%s", shell, stderr.String()))
+			err = errors.Append(err, errors.Wrapf(runErr, "probing shell %q:\n%s", shell, stderr.String()))
 		} else {
 			// Even if there were previous errors, we can now ignore them.
 			err = nil
@@ -629,3 +657,19 @@ func (e stepFailedErr) SingleLineError() string {
 
 	return strings.Split(out, "\n")[0]
 }
+
+type errTimeoutReached struct{ timeout time.Duration }
+
+func (e *errTimeoutReached) Error() string {
+	return fmt.Sprintf("Timeout reached. Execution took longer than %s.", e.timeout)
+}
+
+func reachedTimeout(cmdCtx context.Context, err error) bool {
+	if ee, ok := errors.Cause(err).(*exec.ExitError); ok {
+		if ee.String() == "signal: killed" && cmdCtx.Err() == context.DeadlineExceeded {
+			return true
+		}
+	}
+
+	return errors.Is(errors.Cause(err), context.DeadlineExceeded)
+}
