@@ -4,42 +4,35 @@ import (
 	"context"
 	"encoding/json"
 	"fmt"
-	"html/template"
 	"os"
 	"os/exec"
 	"path"
+	"path/filepath"
 	"strings"
 	"sync"
+	"text/template"
 
 	"github.com/grafana/regexp"
 
-	"github.com/cockroachdb/errors"
-	"github.com/hashicorp/go-multierror"
-
 	batcheslib "github.com/sourcegraph/sourcegraph/lib/batches"
 	onlib "github.com/sourcegraph/sourcegraph/lib/batches/on"
 	templatelib "github.com/sourcegraph/sourcegraph/lib/batches/template"
+	"github.com/sourcegraph/sourcegraph/lib/errors"
 
 	"github.com/sourcegraph/src-cli/internal/api"
 	"github.com/sourcegraph/src-cli/internal/batches"
 	"github.com/sourcegraph/src-cli/internal/batches/docker"
 	"github.com/sourcegraph/src-cli/internal/batches/executor"
 	"github.com/sourcegraph/src-cli/internal/batches/graphql"
-	"github.com/sourcegraph/src-cli/internal/batches/repozip"
 )
 
 type Service struct {
-	allowUnsupported bool
-	allowIgnored     bool
-	client           api.Client
-	features         batches.FeatureFlags
-	imageCache       docker.ImageCache
+	client   api.Client
+	features batches.FeatureFlags
 }
 
 type Opts struct {
-	AllowUnsupported bool
-	AllowIgnored     bool
-	Client           api.Client
+	Client api.Client
 }
 
 var (
@@ -48,10 +41,7 @@ var (
 
 func New(opts *Opts) *Service {
 	return &Service{
-		allowUnsupported: opts.AllowUnsupported,
-		allowIgnored:     opts.AllowIgnored,
-		client:           opts.Client,
-		imageCache:       docker.NewImageCache(),
+		client: opts.Client,
 	}
 }
 
@@ -96,6 +86,10 @@ func (svc *Service) DetermineFeatureFlags(ctx context.Context) error {
 		return errors.Wrap(err, "failed to query Sourcegraph version to check for available features")
 	}
 
+	return svc.SetFeatureFlagsForVersion(version)
+}
+
+func (svc *Service) SetFeatureFlagsForVersion(version string) error {
 	return svc.features.SetFromVersion(version)
 }
 
@@ -161,6 +155,110 @@ func (svc *Service) CreateChangesetSpec(ctx context.Context, spec *batcheslib.Ch
 	return graphql.ChangesetSpecID(result.CreateChangesetSpec.ID), nil
 }
 
+const resolveWorkspacesForBatchSpecQuery = `
+query ResolveWorkspacesForBatchSpec($spec: String!) {
+    resolveWorkspacesForBatchSpec(batchSpec: $spec) {
+		onlyFetchWorkspace
+		ignored
+		unsupported
+		repository {
+			id
+			name
+			url
+			externalRepository { serviceType }
+			defaultBranch {
+				name
+				target { oid }
+			}
+		}
+		branch {
+			name
+			target {
+				oid
+			}
+		}
+		path
+		searchResultPaths
+    }
+}
+`
+
+func (svc *Service) ResolveWorkspacesForBatchSpec(ctx context.Context, spec *batcheslib.BatchSpec, allowUnsupported, allowIgnored bool) ([]RepoWorkspace, []*graphql.Repository, error) {
+	raw, err := json.Marshal(spec)
+	if err != nil {
+		return nil, nil, errors.Wrap(err, "marshalling changeset spec JSON")
+	}
+
+	var result struct {
+		ResolveWorkspacesForBatchSpec []struct {
+			OnlyFetchWorkspace bool
+			Ignored            bool
+			Unsupported        bool
+			Repository         *graphql.Repository
+			Branch             *graphql.Branch
+			Path               string
+			SearchResultPaths  []string
+		}
+	}
+	if ok, err := svc.newRequest(resolveWorkspacesForBatchSpecQuery, map[string]interface{}{
+		"spec": string(raw),
+	}).Do(ctx, &result); err != nil || !ok {
+		return nil, nil, err
+	}
+
+	unsupported := batches.UnsupportedRepoSet{}
+	ignored := batches.IgnoredRepoSet{}
+
+	repos := make([]*graphql.Repository, 0, len(result.ResolveWorkspacesForBatchSpec))
+	seenRepos := make(map[string]struct{})
+	workspaces := make([]RepoWorkspace, 0, len(result.ResolveWorkspacesForBatchSpec))
+	for _, w := range result.ResolveWorkspacesForBatchSpec {
+		fileMatches := make(map[string]bool)
+		for _, path := range w.SearchResultPaths {
+			fileMatches[path] = true
+		}
+
+		workspace := RepoWorkspace{
+			Repo: &graphql.Repository{
+				ID:                 w.Repository.ID,
+				Name:               w.Repository.Name,
+				URL:                w.Repository.URL,
+				FileMatches:        fileMatches,
+				ExternalRepository: w.Repository.ExternalRepository,
+				DefaultBranch:      w.Repository.DefaultBranch,
+				Commit:             w.Branch.Target,
+				Branch:             *w.Branch,
+			},
+			Path:               w.Path,
+			OnlyFetchWorkspace: w.OnlyFetchWorkspace,
+		}
+
+		if !allowIgnored && w.Ignored {
+			ignored.Append(workspace.Repo)
+		} else if !allowUnsupported && w.Unsupported {
+			unsupported.Append(workspace.Repo)
+		} else {
+			// Collect the repo, if not seen yet.
+			if _, ok := seenRepos[workspace.Repo.ID]; !ok {
+				seenRepos[workspace.Repo.ID] = struct{}{}
+				repos = append(repos, workspace.Repo)
+			}
+
+			workspaces = append(workspaces, workspace)
+		}
+	}
+
+	if unsupported.HasUnsupported() {
+		return workspaces, repos, unsupported
+	}
+
+	if ignored.HasIgnored() {
+		return workspaces, repos, ignored
+	}
+
+	return workspaces, repos, nil
+}
+
 // EnsureDockerImages iterates over the steps within the batch spec to ensure the
 // images exist and to determine the exact content digest to be used when running
 // each step, including any required by the service itself.
@@ -168,6 +266,7 @@ func (svc *Service) CreateChangesetSpec(ctx context.Context, spec *batcheslib.Ch
 // Progress information is reported back to the given progress function.
 func (svc *Service) EnsureDockerImages(
 	ctx context.Context,
+	imageCache docker.ImageCache,
 	steps []batcheslib.Step,
 	parallelism int,
 	progress func(done, total int),
@@ -218,7 +317,7 @@ func (svc *Service) EnsureDockerImages(
 					if !more {
 						return
 					}
-					img, err := svc.EnsureImage(workerCtx, name)
+					img, err := imageCache.Ensure(workerCtx, name)
 					select {
 					case <-workerCtx.Done():
 						return
@@ -265,30 +364,16 @@ func (svc *Service) EnsureDockerImages(
 	return images, nil
 }
 
-func (svc *Service) EnsureImage(ctx context.Context, name string) (docker.Image, error) {
-	img := svc.imageCache.Get(name)
-
-	if err := img.Ensure(ctx); err != nil {
-		return nil, errors.Wrapf(err, "pulling image %q", name)
-	}
-
-	return img, nil
-}
-
 func (svc *Service) DetermineWorkspaces(ctx context.Context, repos []*graphql.Repository, spec *batcheslib.BatchSpec) ([]RepoWorkspace, error) {
 	return findWorkspaces(ctx, spec, svc, repos)
 }
 
-func (svc *Service) BuildTasks(ctx context.Context, attributes *templatelib.BatchChangeAttributes, workspaces []RepoWorkspace) []*executor.Task {
-	return buildTasks(ctx, attributes, workspaces)
+func (svc *Service) BuildTasks(attributes *templatelib.BatchChangeAttributes, steps []batcheslib.Step, workspaces []RepoWorkspace) []*executor.Task {
+	return buildTasks(attributes, steps, workspaces)
 }
 
-func (svc *Service) NewCoordinator(opts executor.NewCoordinatorOpts) *executor.Coordinator {
-	opts.RepoArchiveRegistry = repozip.NewArchiveRegistry(svc.client, opts.CacheDir, opts.CleanArchives)
-	opts.Features = svc.features
-	opts.EnsureImage = svc.EnsureImage
-
-	return executor.NewCoordinator(opts)
+func (svc *Service) Features() batches.FeatureFlags {
+	return svc.features
 }
 
 func (svc *Service) CreateImportChangesetSpecs(ctx context.Context, batchSpec *batcheslib.BatchSpec) ([]*batcheslib.ChangesetSpec, error) {
@@ -298,7 +383,7 @@ func (svc *Service) CreateImportChangesetSpecs(ctx context.Context, batchSpec *b
 			repo, err := svc.resolveRepositoryName(ctx, name)
 			if err != nil {
 				wrapped := errors.Wrapf(err, "resolving repository name %q", name)
-				errs = multierror.Append(errs, wrapped)
+				errs = errors.Append(errs, wrapped)
 				continue
 			}
 			repoNameIDs[name] = repo.ID
@@ -373,7 +458,7 @@ func (e *duplicateBranchesErr) Error() string {
 	return out.String()
 }
 
-func (svc *Service) ParseBatchSpec(data []byte) (*batcheslib.BatchSpec, error) {
+func (svc *Service) ParseBatchSpec(dir string, data []byte, isRemote bool) (*batcheslib.BatchSpec, error) {
 	spec, err := batcheslib.ParseBatchSpec(data, batcheslib.ParseBatchSpecOptions{
 		AllowArrayEnvironments: svc.features.AllowArrayEnvironments,
 		AllowTransformChanges:  svc.features.AllowTransformChanges,
@@ -382,9 +467,46 @@ func (svc *Service) ParseBatchSpec(data []byte) (*batcheslib.BatchSpec, error) {
 	if err != nil {
 		return nil, errors.Wrap(err, "parsing batch spec")
 	}
+	if err = validateMount(dir, spec, isRemote); err != nil {
+		return nil, errors.Wrap(err, "handling mount")
+	}
 	return spec, nil
 }
 
+func validateMount(batchSpecDir string, spec *batcheslib.BatchSpec, isRemote bool) error {
+	for i, step := range spec.Steps {
+		for j, mount := range step.Mount {
+			if isRemote {
+				return errors.New("mounts are not supported for server-side processing")
+			}
+			p := mount.Path
+			if !filepath.IsAbs(p) {
+				// Try to build the absolute path since Docker will only mount absolute paths
+				p = filepath.Join(batchSpecDir, p)
+			}
+			pathInfo, err := os.Stat(p)
+			if os.IsNotExist(err) {
+				return errors.Newf("step %d mount path %s does not exist", i+1, p)
+			} else if err != nil {
+				return errors.Wrapf(err, "step %d mount path validation", i+1)
+			}
+			if !strings.HasPrefix(p, batchSpecDir) {
+				return errors.Newf("step %d mount path is not in the same directory or subdirectory as the batch spec", i+1)
+			}
+			// Mounting a directory on Docker must end with the separator. So, append the file separator to make
+			// users' lives easier.
+			if pathInfo.IsDir() && !strings.HasSuffix(p, string(filepath.Separator)) {
+				p += string(filepath.Separator)
+			}
+			// Update the mount path to the absolute path so building the absolute path (above) does not need to be
+			// redone when adding the mount argument to the Docker container.
+			// TODO: Can this mess with caching? We wouldn't be doing that server-side.
+			step.Mount[j].Path = p
+		}
+	}
+	return nil
+}
+
 const exampleSpecTmpl = `name: NAME-OF-YOUR-BATCH-CHANGE
 description: DESCRIPTION-OF-YOUR-BATCH-CHANGE
 
@@ -465,10 +587,12 @@ const namespaceQuery = `
 query NamespaceQuery($name: String!) {
     user(username: $name) {
         id
+        url
     }
 
     organization(name: $name) {
         id
+        url
     }
 }
 `
@@ -477,53 +601,75 @@ const usernameQuery = `
 query GetCurrentUserID {
     currentUser {
         id
+        url
     }
 }
 `
 
-func (svc *Service) ResolveNamespace(ctx context.Context, namespace string) (string, error) {
+type Namespace struct {
+	ID  string
+	URL string
+}
+
+func (svc *Service) ResolveNamespace(ctx context.Context, namespace string) (Namespace, error) {
 	if namespace == "" {
 		// if no namespace is provided, default to logged in user as namespace
 		var resp struct {
 			Data struct {
 				CurrentUser struct {
-					ID string `json:"id"`
+					ID  string `json:"id"`
+					URL string `json:"url"`
 				} `json:"currentUser"`
 			} `json:"data"`
 		}
 		if ok, err := svc.client.NewRequest(usernameQuery, nil).DoRaw(ctx, &resp); err != nil || !ok {
-			return "", errors.WithMessage(err, "failed to resolve namespace: no user logged in")
+			return Namespace{}, errors.WithMessage(err, "failed to resolve namespace: no user logged in")
 		}
 
 		if resp.Data.CurrentUser.ID == "" {
-			return "", errors.New("cannot resolve current user")
+			return Namespace{}, errors.New("cannot resolve current user")
 		}
-		return resp.Data.CurrentUser.ID, nil
+		return Namespace{
+			ID:  resp.Data.CurrentUser.ID,
+			URL: resp.Data.CurrentUser.URL,
+		}, nil
 	}
 
 	var result struct {
 		Data struct {
-			User         *struct{ ID string }
-			Organization *struct{ ID string }
+			User *struct {
+				ID  string
+				URL string
+			}
+			Organization *struct {
+				ID  string
+				URL string
+			}
 		}
 		Errors []interface{}
 	}
 	if ok, err := svc.client.NewRequest(namespaceQuery, map[string]interface{}{
 		"name": namespace,
 	}).DoRaw(ctx, &result); err != nil || !ok {
-		return "", err
+		return Namespace{}, err
 	}
 
 	if result.Data.User != nil {
-		return result.Data.User.ID, nil
+		return Namespace{
+			ID:  result.Data.User.ID,
+			URL: result.Data.User.URL,
+		}, nil
 	}
 	if result.Data.Organization != nil {
-		return result.Data.Organization.ID, nil
+		return Namespace{
+			ID:  result.Data.Organization.ID,
+			URL: result.Data.Organization.URL,
+		}, nil
 	}
-	return "", fmt.Errorf("failed to resolve namespace %q: no user or organization found", namespace)
+	return Namespace{}, fmt.Errorf("failed to resolve namespace %q: no user or organization found", namespace)
 }
 
-func (svc *Service) ResolveRepositories(ctx context.Context, spec *batcheslib.BatchSpec) ([]*graphql.Repository, error) {
+func (svc *Service) ResolveRepositories(ctx context.Context, spec *batcheslib.BatchSpec, allowUnsupported, allowIgnored bool) ([]*graphql.Repository, error) {
 	agg := onlib.NewRepoRevisionAggregator()
 	unsupported := batches.UnsupportedRepoSet{}
 	ignored := batches.IgnoredRepoSet{}
@@ -532,7 +678,7 @@ func (svc *Service) ResolveRepositories(ctx context.Context, spec *batcheslib.Ba
 	for _, on := range spec.On {
 		repos, ruleType, err := svc.ResolveRepositoriesOn(ctx, &on)
 		if err != nil {
-			return nil, errors.Wrapf(err, "resolving %q", on.String())
+			return nil, errors.Wrapf(err, on.String())
 		}
 
 		result := agg.NewRuleRevisions(ruleType)
@@ -545,7 +691,7 @@ func (svc *Service) ResolveRepositories(ctx context.Context, spec *batcheslib.Ba
 		}
 
 		var repoBatchIgnores map[*graphql.Repository][]string
-		if !svc.allowIgnored {
+		if !allowIgnored {
 			repoBatchIgnores, err = svc.FindDirectoriesInRepos(ctx, ".batchignore", reposWithBranch...)
 			if err != nil {
 				return nil, err
@@ -557,13 +703,18 @@ func (svc *Service) ResolveRepositories(ctx context.Context, spec *batcheslib.Ba
 
 			switch st := strings.ToLower(repo.ExternalRepository.ServiceType); st {
 			case "github", "gitlab", "bitbucketserver":
+			case "bitbucketcloud":
+				if svc.features.BitbucketCloud {
+					break
+				}
+				fallthrough
 			default:
-				if !svc.allowUnsupported {
+				if !allowUnsupported {
 					unsupported.Append(repo)
 				}
 			}
 
-			if !svc.allowIgnored {
+			if !allowIgnored {
 				if locations, ok := repoBatchIgnores[repo]; ok && len(locations) > 0 {
 					ignored.Append(repo)
 				}
@@ -644,7 +795,7 @@ func (svc *Service) resolveRepositoryName(ctx context.Context, name string) (*gr
 		return nil, err
 	}
 	if result.Repository == nil {
-		return nil, errors.New("no repository found")
+		return nil, errors.New("no repository found: check spelling and specify the repository in the format \"<codehost_url>/owner/repo-name\" or \"repo-name\" as required by your instance")
 	}
 	return result.Repository, nil
 }
