@@ -6,14 +6,18 @@ import (
 	"fmt"
 	"net/http"
 	"net/http/httptest"
+	"os"
+	"path/filepath"
 	"strconv"
 	"strings"
 	"testing"
 
-	"github.com/cockroachdb/errors"
 	"github.com/google/go-cmp/cmp"
 	"github.com/google/go-cmp/cmp/cmpopts"
 	"github.com/stretchr/testify/assert"
+	"github.com/stretchr/testify/require"
+
+	"github.com/sourcegraph/sourcegraph/lib/errors"
 
 	batcheslib "github.com/sourcegraph/sourcegraph/lib/batches"
 
@@ -130,7 +134,7 @@ func mockGraphQLClient(responses ...string) (client api.Client, done func()) {
 	mux := http.NewServeMux()
 
 	var count int
-	mux.HandleFunc("/.api/graphql", func(w http.ResponseWriter, r *http.Request) {
+	mux.HandleFunc("/.api/graphql", func(w http.ResponseWriter, _ *http.Request) {
 		w.Header().Set("Content-Type", "application/json")
 
 		_, _ = w.Write([]byte(responses[count]))
@@ -159,9 +163,9 @@ func TestResolveRepositories_Unsupported(t *testing.T) {
 	}
 
 	t.Run("allowUnsupported:true", func(t *testing.T) {
-		svc := &Service{client: client, allowUnsupported: true, allowIgnored: true}
+		svc := &Service{client: client}
 
-		repos, err := svc.ResolveRepositories(context.Background(), spec)
+		repos, err := svc.ResolveRepositories(context.Background(), spec, true, true)
 		if err != nil {
 			t.Fatalf("unexpected error: %s", err)
 		}
@@ -171,9 +175,9 @@ func TestResolveRepositories_Unsupported(t *testing.T) {
 	})
 
 	t.Run("allowUnsupported:false", func(t *testing.T) {
-		svc := &Service{client: client, allowUnsupported: false, allowIgnored: true}
+		svc := &Service{client: client}
 
-		repos, err := svc.ResolveRepositories(context.Background(), spec)
+		repos, err := svc.ResolveRepositories(context.Background(), spec, false, true)
 		repoSet, ok := err.(batches.UnsupportedRepoSet)
 		if !ok {
 			t.Fatalf("err is not UnsupportedRepoSet")
@@ -242,9 +246,9 @@ func TestResolveRepositories_Ignored(t *testing.T) {
 		client, done := mockGraphQLClient(testResolveRepositories, testBatchIgnoreInRepos)
 		defer done()
 
-		svc := &Service{client: client, allowIgnored: true}
+		svc := &Service{client: client}
 
-		repos, err := svc.ResolveRepositories(context.Background(), spec)
+		repos, err := svc.ResolveRepositories(context.Background(), spec, false, true)
 		if err != nil {
 			t.Fatalf("unexpected error: %s", err)
 		}
@@ -257,9 +261,9 @@ func TestResolveRepositories_Ignored(t *testing.T) {
 		client, done := mockGraphQLClient(testResolveRepositories, testBatchIgnoreInRepos)
 		defer done()
 
-		svc := &Service{client: client, allowIgnored: false}
+		svc := &Service{client: client}
 
-		repos, err := svc.ResolveRepositories(context.Background(), spec)
+		repos, err := svc.ResolveRepositories(context.Background(), spec, false, false)
 		ignored, ok := err.(batches.IgnoredRepoSet)
 		if !ok {
 			t.Fatalf("err is not IgnoredRepoSet: %s", err)
@@ -339,9 +343,9 @@ func TestResolveRepositories_RepoWithoutBranch(t *testing.T) {
 	client, done := mockGraphQLClient(testResolveRepositoriesNoBranch, testBatchIgnoreInReposNoBranch)
 	defer done()
 
-	svc := &Service{client: client, allowIgnored: false}
+	svc := &Service{client: client}
 
-	repos, err := svc.ResolveRepositories(context.Background(), spec)
+	repos, err := svc.ResolveRepositories(context.Background(), spec, false, false)
 	if err != nil {
 		t.Fatalf("unexpected error: %s", err)
 	}
@@ -580,11 +584,7 @@ func TestEnsureDockerImages(t *testing.T) {
 	ctx := context.Background()
 	parallelCases := []int{0, 1, 2, 4, 8}
 
-	newServiceWithImages := func(images map[string]docker.Image) *Service {
-		return &Service{
-			imageCache: &mock.ImageCache{Images: images},
-		}
-	}
+	svc := &Service{}
 
 	t.Run("success", func(t *testing.T) {
 		t.Run("single image", func(t *testing.T) {
@@ -601,10 +601,8 @@ func TestEnsureDockerImages(t *testing.T) {
 				t.Run(name, func(t *testing.T) {
 					for _, parallelism := range parallelCases {
 						t.Run(fmt.Sprintf("%d worker(s)", parallelism), func(t *testing.T) {
-							svc := newServiceWithImages(images)
 							progress := &mock.Progress{}
-
-							have, err := svc.EnsureDockerImages(ctx, steps, parallelism, progress.Callback())
+							have, err := svc.EnsureDockerImages(ctx, &mock.ImageCache{Images: images}, steps, parallelism, progress.Callback())
 							assert.Nil(t, err)
 							assert.Equal(t, images, have)
 							assert.Equal(t, []mock.ProgressCall{
@@ -631,16 +629,21 @@ func TestEnsureDockerImages(t *testing.T) {
 
 			for _, parallelism := range parallelCases {
 				t.Run(fmt.Sprintf("%d worker(s)", parallelism), func(t *testing.T) {
-					svc := newServiceWithImages(images)
 					progress := &mock.Progress{}
 
-					have, err := svc.EnsureDockerImages(ctx, []batcheslib.Step{
-						{Container: "a"},
-						{Container: "a"},
-						{Container: "a"},
-						{Container: "b"},
-						{Container: "c"},
-					}, parallelism, progress.Callback())
+					have, err := svc.EnsureDockerImages(
+						ctx,
+						&mock.ImageCache{Images: images},
+						[]batcheslib.Step{
+							{Container: "a"},
+							{Container: "a"},
+							{Container: "a"},
+							{Container: "b"},
+							{Container: "c"},
+						},
+						parallelism,
+						progress.Callback(),
+					)
 					assert.Nil(t, err)
 					assert.Equal(t, images, have)
 					assert.Equal(t, []mock.ProgressCall{
@@ -682,10 +685,9 @@ func TestEnsureDockerImages(t *testing.T) {
 
 		for _, parallelism := range parallelCases {
 			t.Run(fmt.Sprintf("%d worker(s)", parallelism), func(t *testing.T) {
-				svc := newServiceWithImages(images)
 				progress := &mock.Progress{}
 
-				have, err := svc.EnsureDockerImages(ctx, steps, parallelism, progress.Callback())
+				have, err := svc.EnsureDockerImages(ctx, &mock.ImageCache{Images: images}, steps, parallelism, progress.Callback())
 				assert.ErrorIs(t, err, wantErr)
 				assert.Nil(t, have)
 
@@ -698,3 +700,343 @@ func TestEnsureDockerImages(t *testing.T) {
 
 	})
 }
+
+func TestService_ParseBatchSpec(t *testing.T) {
+	svc := &Service{}
+
+	tempDir := t.TempDir()
+	tempOutsideDir := t.TempDir()
+	// create temp files/dirs that can be used by the tests
+	_, err := os.Create(filepath.Join(tempDir, "sample.sh"))
+	require.NoError(t, err)
+	_, err = os.Create(filepath.Join(tempDir, "another.sh"))
+	require.NoError(t, err)
+
+	tests := []struct {
+		name         string
+		batchSpecDir string
+		rawSpec      string
+		isRemote     bool
+		expectedSpec *batcheslib.BatchSpec
+		expectedErr  error
+	}{
+		{
+			name: "simple spec",
+			rawSpec: `
+name: test-spec
+description: A test spec
+`,
+			expectedSpec: &batcheslib.BatchSpec{Name: "test-spec", Description: "A test spec"},
+		},
+		{
+			name: "unknown field",
+			rawSpec: `
+name: test-spec
+description: A test spec
+some-new-field: Foo bar
+`,
+			expectedErr: errors.New("parsing batch spec: Additional property some-new-field is not allowed"),
+		},
+		{
+			name:         "mount absolute file",
+			batchSpecDir: tempDir,
+			rawSpec: fmt.Sprintf(`
+name: test-spec
+description: A test spec
+steps:
+  - run: /tmp/sample.sh
+    container: alpine:3
+    mount:
+      - path: %s
+        mountpoint: /tmp/sample.sh
+changesetTemplate:
+  title: Test Mount
+  body: Test a mounted path
+  branch: test
+  commit:
+    message: Test
+`, filepath.Join(tempDir, "sample.sh")),
+			expectedSpec: &batcheslib.BatchSpec{
+				Name:        "test-spec",
+				Description: "A test spec",
+				Steps: []batcheslib.Step{
+					{
+						Run:       "/tmp/sample.sh",
+						Container: "alpine:3",
+						Mount: []batcheslib.Mount{
+							{
+								Path:       filepath.Join(tempDir, "sample.sh"),
+								Mountpoint: "/tmp/sample.sh",
+							},
+						},
+					},
+				},
+				ChangesetTemplate: &batcheslib.ChangesetTemplate{
+					Title:  "Test Mount",
+					Body:   "Test a mounted path",
+					Branch: "test",
+					Commit: batcheslib.ExpandedGitCommitDescription{
+						Message: "Test",
+					},
+				},
+			},
+		},
+		{
+			name:         "mount relative file",
+			batchSpecDir: tempDir,
+			rawSpec: `
+name: test-spec
+description: A test spec
+steps:
+  - run: /tmp/sample.sh
+    container: alpine:3
+    mount:
+      - path: ./sample.sh
+        mountpoint: /tmp/sample.sh
+changesetTemplate:
+  title: Test Mount
+  body: Test a mounted path
+  branch: test
+  commit:
+    message: Test
+`,
+			expectedSpec: &batcheslib.BatchSpec{
+				Name:        "test-spec",
+				Description: "A test spec",
+				Steps: []batcheslib.Step{
+					{
+						Run:       "/tmp/sample.sh",
+						Container: "alpine:3",
+						Mount: []batcheslib.Mount{
+							{
+								Path:       filepath.Join(tempDir, "sample.sh"),
+								Mountpoint: "/tmp/sample.sh",
+							},
+						},
+					},
+				},
+				ChangesetTemplate: &batcheslib.ChangesetTemplate{
+					Title:  "Test Mount",
+					Body:   "Test a mounted path",
+					Branch: "test",
+					Commit: batcheslib.ExpandedGitCommitDescription{
+						Message: "Test",
+					},
+				},
+			},
+		},
+		{
+			name:         "mount absolute directory",
+			batchSpecDir: tempDir,
+			rawSpec: fmt.Sprintf(`
+name: test-spec
+description: A test spec
+steps:
+  - run: /tmp/some-file.sh
+    container: alpine:3
+    mount:
+      - path: %s
+        mountpoint: /tmp
+changesetTemplate:
+  title: Test Mount
+  body: Test a mounted path
+  branch: test
+  commit:
+    message: Test
+`, tempDir),
+			expectedSpec: &batcheslib.BatchSpec{
+				Name:        "test-spec",
+				Description: "A test spec",
+				Steps: []batcheslib.Step{
+					{
+						Run:       "/tmp/some-file.sh",
+						Container: "alpine:3",
+						Mount: []batcheslib.Mount{
+							{
+								Path:       tempDir + string(filepath.Separator),
+								Mountpoint: "/tmp",
+							},
+						},
+					},
+				},
+				ChangesetTemplate: &batcheslib.ChangesetTemplate{
+					Title:  "Test Mount",
+					Body:   "Test a mounted path",
+					Branch: "test",
+					Commit: batcheslib.ExpandedGitCommitDescription{
+						Message: "Test",
+					},
+				},
+			},
+		},
+		{
+			name:         "mount relative directory",
+			batchSpecDir: tempDir,
+			rawSpec: `
+name: test-spec
+description: A test spec
+steps:
+  - run: /tmp/some-file.sh
+    container: alpine:3
+    mount:
+      - path: ./
+        mountpoint: /tmp
+changesetTemplate:
+  title: Test Mount
+  body: Test a mounted path
+  branch: test
+  commit:
+    message: Test
+`,
+			expectedSpec: &batcheslib.BatchSpec{
+				Name:        "test-spec",
+				Description: "A test spec",
+				Steps: []batcheslib.Step{
+					{
+						Run:       "/tmp/some-file.sh",
+						Container: "alpine:3",
+						Mount: []batcheslib.Mount{
+							{
+								Path:       tempDir + string(filepath.Separator),
+								Mountpoint: "/tmp",
+							},
+						},
+					},
+				},
+				ChangesetTemplate: &batcheslib.ChangesetTemplate{
+					Title:  "Test Mount",
+					Body:   "Test a mounted path",
+					Branch: "test",
+					Commit: batcheslib.ExpandedGitCommitDescription{
+						Message: "Test",
+					},
+				},
+			},
+		},
+		{
+			name:         "mount multiple files",
+			batchSpecDir: tempDir,
+			rawSpec: `
+name: test-spec
+description: A test spec
+steps:
+  - run: /tmp/sample.sh && /tmp/another.sh
+    container: alpine:3
+    mount:
+      - path: ./sample.sh
+        mountpoint: /tmp/sample.sh
+      - path: ./another.sh
+        mountpoint: /tmp/another.sh
+changesetTemplate:
+  title: Test Mount
+  body: Test a mounted path
+  branch: test
+  commit:
+    message: Test
+`,
+			expectedSpec: &batcheslib.BatchSpec{
+				Name:        "test-spec",
+				Description: "A test spec",
+				Steps: []batcheslib.Step{
+					{
+						Run:       "/tmp/sample.sh && /tmp/another.sh",
+						Container: "alpine:3",
+						Mount: []batcheslib.Mount{
+							{
+								Path:       filepath.Join(tempDir, "sample.sh"),
+								Mountpoint: "/tmp/sample.sh",
+							},
+							{
+								Path:       filepath.Join(tempDir, "another.sh"),
+								Mountpoint: "/tmp/another.sh",
+							},
+						},
+					},
+				},
+				ChangesetTemplate: &batcheslib.ChangesetTemplate{
+					Title:  "Test Mount",
+					Body:   "Test a mounted path",
+					Branch: "test",
+					Commit: batcheslib.ExpandedGitCommitDescription{
+						Message: "Test",
+					},
+				},
+			},
+		},
+		{
+			name:         "mount path does not exist",
+			batchSpecDir: tempDir,
+			rawSpec: fmt.Sprintf(`
+name: test-spec
+description: A test spec
+steps:
+  - run: /tmp/sample.sh
+    container: alpine:3
+    mount:
+      - path: %s
+        mountpoint: /tmp
+changesetTemplate:
+  title: Test Mount
+  body: Test a mounted path
+  branch: test
+  commit:
+    message: Test
+`, filepath.Join(tempDir, "does", "not", "exist", "sample.sh")),
+			expectedErr: errors.Newf("handling mount: step 1 mount path %s does not exist", filepath.Join(tempDir, "does", "not", "exist", "sample.sh")),
+		},
+		{
+			name:         "mount path not subdirectory of spec",
+			batchSpecDir: tempDir,
+			rawSpec: fmt.Sprintf(`
+name: test-spec
+description: A test spec
+steps:
+  - run: /tmp/sample.sh
+    container: alpine:3
+    mount:
+      - path: %s
+        mountpoint: /tmp
+changesetTemplate:
+  title: Test Mount
+  body: Test a mounted path
+  branch: test
+  commit:
+    message: Test
+`, tempOutsideDir),
+			expectedErr: errors.New("handling mount: step 1 mount path is not in the same directory or subdirectory as the batch spec"),
+		},
+		{
+			name:         "mount remote processing",
+			batchSpecDir: tempDir,
+			rawSpec: `
+name: test-spec
+description: A test spec
+steps:
+  - run: /tmp/foo/bar/sample.sh
+    container: alpine:3
+    mount:
+      - path: /valid/sample.sh
+        mountpoint: /tmp/foo/bar/sample.sh
+changesetTemplate:
+  title: Test Mount
+  body: Test a mounted path
+  branch: test
+  commit:
+    message: Test
+`,
+			isRemote:    true,
+			expectedErr: errors.New("handling mount: mounts are not supported for server-side processing"),
+		},
+	}
+	for _, test := range tests {
+		t.Run(test.name, func(t *testing.T) {
+			spec, err := svc.ParseBatchSpec(test.batchSpecDir, []byte(test.rawSpec), test.isRemote)
+			if test.expectedErr != nil {
+				assert.Equal(t, test.expectedErr.Error(), err.Error())
+			} else {
+				assert.NoError(t, err)
+				assert.Equal(t, test.expectedSpec, spec)
+			}
+		})
+	}
+}
